<?php

declare(strict_types=1);

namespace App;

use Dotenv\Dotenv;
use GuzzleHttp\Client;
use GuzzleHttp\Cookie\CookieJar;
use GuzzleHttp\Cookie\SetCookie;
use GuzzleHttp\RequestOptions;
use Symfony\Component\Console\Attribute\AsCommand;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Webmozart\Assert\Assert;
use function Functional\map;
use function Functional\partition;

#[AsCommand(
    name: 'cards:exchange',
    description: 'Get and exchange cards',
)]
class ExchangeCardsCommand extends Command
{
    private const CARDS = [
        144, // Удачный поворот: +4 карты
        145, // Колесо Фортуны: +12 карт
        146, // Белая полоса: +42 карты
        147, // Прядь Кайроса: +150 карт
    ];

    private readonly string $csrfToken;
    private readonly CookieJar $cookies;
    private readonly Client $client;

    public function __construct(string $name = null)
    {
        parent::__construct($name);

        $dotenv = Dotenv::createImmutable(__DIR__ . '/../');
        $dotenv->load();

        $this->csrfToken = bin2hex(random_bytes(32));

        $this->cookies = new CookieJar();
        $this->cookies->setCookie(new SetCookie([
            'Name' => 'csrftoken',
            'Value' => $this->csrfToken,
            'Domain' => 'the-tale.org',
        ]));

        $this->client = new Client();
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        $this->login();

        $output->writeln('Receive new cards');
        $this->receiveNewCards();

        do {
            $output->writeln('Fetch cards');
            $cards = $this->getCards();
            $output->writeln('Find special cards');
            $specialCards = array_filter($cards, static fn (Card $c) => in_array($c->type, self::CARDS));
            if (count($specialCards) > 0) {
                foreach ($specialCards as $card) {
                    $output->writeln('Use card ' . $card->name);
                    $this->useCard($card->uid);
                }
            }

            $output->writeln('Wait for cards applies');
            sleep(3);

            $output->writeln('Receive new cards');
            $this->receiveNewCards();
            $cards = $this->getCards();

            $output->writeln('Combine cards');
            $combined = 0;
            foreach ([0, 1, 2, 3] as $rarity) {
                $selectedCards = array_filter($cards, static fn (Card $c) => $c->rarity === $rarity && !in_array($c->type, self::CARDS));
                while (true) {
                    [$next, $rest] = partition(array_values($selectedCards), fn ($_, $index) => $index < 3);
                    if (count($next) < 3) {
                        break;
                    }
                    $this->combileCards(map($next, fn (Card $c) => $c->uid));
                    $selectedCards = $rest;
                    $combined++;
                }
            }
            $output->writeln('Combined cards: ' . $combined);
        } while ($combined !== 0);

        return self::SUCCESS;
    }

    private function login()
    {
        $response = $this->client->request('POST', 'https://the-tale.org/accounts/auth/api/login', [
            RequestOptions::HEADERS => [
                'Referer' => 'https://the-tale.org/',
                'X-CSRFToken' => $this->csrfToken,
            ],
            RequestOptions::QUERY => [
                'api_version' => '1.0',
                'api_client' => 'anwinged-1',
            ],
            RequestOptions::FORM_PARAMS => [
                'email' => (string) $_ENV['TALE_EMAIL'],
                'password' => (string) $_ENV['TALE_PASS'],
            ],
            RequestOptions::COOKIES => $this->cookies,
        ]);

        Assert::eq($response->getStatusCode(), 200);
    }

    /**
     * @return Card[]
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    private function getCards(): array
    {
        $response = $this->client->request('GET', 'https://the-tale.org/game/cards/api/get-cards', [
            RequestOptions::HEADERS => [
                'Referer' => 'https://the-tale.org/',
            ],
            RequestOptions::QUERY => [
                'api_version' => '2.0',
                'api_client' => 'anwinged-1',
            ],
            RequestOptions::COOKIES => $this->cookies,
        ]);

        Assert::eq($response->getStatusCode(), 200);

        $data = json_decode((string) $response->getBody(), associative: true);

        $cardsArray = $data['data']['cards'];
        Assert::isArray($cardsArray);

        return array_map(Card::fromArray(...), $cardsArray);
    }

    private function receiveNewCards()
    {
        $response = $this->client->request('POST', 'https://the-tale.org/game/cards/api/receive', [
            RequestOptions::HEADERS => [
                'Referer' => 'https://the-tale.org/',
                'X-CSRFToken' => $this->cookies->getCookieByName('csrftoken')?->getValue(),
            ],
            RequestOptions::QUERY => [
                'api_version' => '1.0',
                'api_client' => 'anwinged-1',
            ],
            RequestOptions::COOKIES => $this->cookies,
        ]);

        Assert::eq($response->getStatusCode(), 200);
    }

    private function useCard(string $uid)
    {
        $response = $this->client->request('POST', 'https://the-tale.org/game/cards/api/use', [
            RequestOptions::HEADERS => [
                'Referer' => 'https://the-tale.org/',
                'X-CSRFToken' => $this->cookies->getCookieByName('csrftoken')?->getValue(),
            ],
            RequestOptions::QUERY => [
                'api_version' => '2.0',
                'api_client' => 'anwinged-1',
                'card' => $uid,
            ],
            RequestOptions::COOKIES => $this->cookies,
        ]);

        Assert::eq($response->getStatusCode(), 200);

        dump((string) $response->getBody());
    }

    private function combileCards(array $uids)
    {
        $cardsData = implode('&', map($uids, fn ($uid) => 'card=' . $uid));

        $response = $this->client->request('POST', 'https://the-tale.org/game/cards/api/combine', [
            RequestOptions::HEADERS => [
                'Referer' => 'https://the-tale.org/',
                'X-CSRFToken' => $this->cookies->getCookieByName('csrftoken')?->getValue(),
                'Content-Type' => 'application/x-www-form-urlencoded',
            ],
            RequestOptions::QUERY => [
                'api_version' => '3.0',
                'api_client' => 'anwinged-1',
            ],
            RequestOptions::BODY => $cardsData,
            RequestOptions::COOKIES => $this->cookies,
        ]);

        Assert::eq($response->getStatusCode(), 200);

        dump((string) $response->getBody());
    }
}
